package iotpolicy;

import java_cup.runtime.ComplexSymbolFactory;
import java_cup.runtime.ScannerBuffer;
import java.io.*;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import iotpolicy.parser.Lexer;
import iotpolicy.parser.Parser;
import iotpolicy.tree.ParseNode;
import iotpolicy.tree.ParseNodeVector;
import iotpolicy.tree.ParseTreeHandler;
import iotpolicy.tree.Declaration;
import iotpolicy.tree.DeclarationHandler;
import iotpolicy.tree.CapabilityDecl;
import iotpolicy.tree.InterfaceDecl;
import iotpolicy.tree.RequiresDecl;


/** Class IoTCompiler is the main interface/stub compiler for
 *  files generation. This class calls helper classes
 *  such as Parser, Lexer, InterfaceDecl, CapabilityDecl,
 *  RequiresDecl, ParseTreeHandler, etc.
 *
 * @author      Rahmadi Trimananda <rahmadi.trimananda @ uci.edu>
 * @version     1.0
 * @since       2016-09-22
 */
public class IoTCompiler {

	/**
	 * Class properties
	 */
	// Maps multiple interfaces to multiple objects of ParseTreeHandler
	private Map<String,ParseTreeHandler> mapIntfacePTH;
	private Map<String,DeclarationHandler> mapIntDeclHand;
	private Map<String,Map<String,Set<String>>> mapInt2NewInts;
	// Data structure to store our types (primitives and non-primitives) for compilation
	private Map<String,String> mapPrimitives;
	private Map<String,String> mapNonPrimitivesJava;
	private Map<String,String> mapNonPrimitivesCplus;
	private PrintWriter pw;
	private String dir;
	private String subdir;

	/**
	 * Class constants
	 */
	private final static String OUTPUT_DIRECTORY = "output_files";

	/**
	 * Primitive data types
	 */
	private final static String[] primitives = new String[] {

		"byte",
		"Byte",
		"short",
		"Short",
		"int",
		"Integer",
		"long",
		"Long",
		"float",
		"Float",
		"double",
		"Double",
		"boolean",
		"Boolean",
		"char",
		"Character",
		"string",
		"String",
		"void"
	};

	/**
	 * Primitive data types in C++ to map the primitives list
	 */
	private final static String[] primitivesCplus = new String[] {

		"char",
		"char",
		"short",
		"short",
		"int",
		"int",
		"long",
		"long",
		"float",
		"float",
		"double",
		"double",
		"bool",
		"bool",
		"char",
		"char",
		"string",
		"string",
		"void"
	};

	/**
	 * Non-primitive data types supported by this compiler
	 */
	private final static String[] nonPrimitives = new String[] {

		"Set",
		"HashSet",
		"Map",
		"HashMap",
		"List",
		"ArrayList"
	};

	/**
	 * Non-primitive Java libraries based on the list above
	 */
	private final static String[] nonPrimitiveJavaLibs = new String[] {

		"java.util.Set",
		"java.util.HashSet",
		"java.util.Map",
		"java.util.HashMap",
		"java.util.List",
		"java.util.ArrayList"
	};

	/**
	 * Non-primitive C++ libraries based on the list above
	 */
	private final static String[] nonPrimitiveCplusLibs = new String[] {

		"set",
		"unordered_set",
		"map",
		"unordered_map",
		"list",
		"list"
	};

	private enum ParamCategory {

		PRIMITIVES,		// All the primitive types, e.g. byte, short, int, long, etc.
		NONPRIMITIVES,	// Non-primitive types, e.g. Set, Map, List, etc.
		USERDEFINED		// Non-supported type by default; assumed as driver classes
	}

	/**
	 * Class constructors
	 */
	public IoTCompiler() {

		mapIntfacePTH = new HashMap<String,ParseTreeHandler>();
		mapIntDeclHand = new HashMap<String,DeclarationHandler>();
		mapInt2NewInts = new HashMap<String,Map<String,Set<String>>>();
		mapPrimitives = new HashMap<String,String>();
			arraysToMap(mapPrimitives, primitives, primitivesCplus);
		mapNonPrimitivesJava = new HashMap<String,String>();
			arraysToMap(mapNonPrimitivesJava, nonPrimitives, nonPrimitiveJavaLibs);
		mapNonPrimitivesCplus = new HashMap<String,String>();
			arraysToMap(mapNonPrimitivesCplus, nonPrimitives, nonPrimitiveCplusLibs);
		pw = null;
		dir = OUTPUT_DIRECTORY;
		subdir = null;
	}


	/**
	 * setParseTree() sets parse tree based on policy files.
	 * <p>
	 * It also generates parse tree (ParseTreeHandler) and
	 * copies useful information from parse tree into
	 * InterfaceDecl, CapabilityDecl, and RequiresDecl 
	 * data structures.
	 * Additionally, the data structure handles are
	 * returned from tree-parsing for further process.
	 *
	 */
	public void setParseTree(String origInt, ParseNode pnPol, ParseNode pnReq) {

		ParseTreeHandler ptHandler = new ParseTreeHandler(origInt, pnPol, pnReq);
		DeclarationHandler decHandler = new DeclarationHandler();

		// Process ParseNode and generate Declaration objects
		ptHandler.processInterfaceDecl();
		InterfaceDecl intDecl = ptHandler.getInterfaceDecl();
		decHandler.addInterfaceDecl(origInt, intDecl);
		ptHandler.processCapabilityDecl();
		CapabilityDecl capDecl = ptHandler.getCapabilityDecl();
		decHandler.addCapabilityDecl(origInt, capDecl);
		ptHandler.processRequiresDecl();
		RequiresDecl reqDecl = ptHandler.getRequiresDecl();
		decHandler.addRequiresDecl(origInt, reqDecl);

		mapIntfacePTH.put(origInt, ptHandler);
		mapIntDeclHand.put(origInt, decHandler);
	}


	/**
	 * getMethodsForIntface() reads for methods in the data structure
	 * <p>
	 * It is going to give list of methods for a certain interface
	 * 		based on the declaration of capabilities.
	 */
	public void getMethodsForIntface(String origInt) {

		ParseTreeHandler ptHandler = mapIntfacePTH.get(origInt);
		Map<String,Set<String>> mapNewIntMethods = new HashMap<String,Set<String>>();
		// Get set of new interfaces, e.g. CameraWithCaptureAndData
		// Generate this new interface with all the methods it needs
		//		from different capabilities it declares
		DeclarationHandler decHandler = mapIntDeclHand.get(origInt);
		RequiresDecl reqDecl = (RequiresDecl) decHandler.getRequiresDecl(origInt);
		Set<String> setIntfaces = reqDecl.getInterfaces();
		for (String strInt : setIntfaces) {

			// Initialize a set of methods
			Set<String> setMethods = new HashSet<String>();
			// Get list of capabilities, e.g. ImageCapture, VideoRecording, etc.
			List<String> listCapab = reqDecl.getCapabList(strInt);
			for (String strCap : listCapab) {

				// Get list of methods for each capability
				CapabilityDecl capDecl = (CapabilityDecl) decHandler.getCapabilityDecl(origInt);
				List<String> listCapabMeth = capDecl.getMethods(strCap);
				for (String strMeth : listCapabMeth) {

					// Add methods into setMethods
					// This is to also handle redundancies (say two capabilities
					//		share the same methods)
					setMethods.add(strMeth);
				}
			}
			// Add interface and methods information into map
			mapNewIntMethods.put(strInt, setMethods);
		}
		// Map the map of interface-methods to the original interface
		mapInt2NewInts.put(origInt, mapNewIntMethods);
	}


	/**
	 * generateJavaLocalInterface() writes the local interface and provides type-checking.
	 * <p>
	 * It needs to rewrite and exchange USERDEFINED types in input parameters of stub
	 * and original interfaces, e.g. exchange Camera and CameraWithVideoAndRecording.
	 * The local interface has to be the input parameter for the stub and the stub 
	 * interface has to be the input parameter for the local class.
	 */
	public void generateJavaLocalInterfaces() throws IOException {

		// Create a new directory
		createDirectory(dir);
		for (String intface : mapIntfacePTH.keySet()) {
			// Open a new file to write into
			FileWriter fw = new FileWriter(dir + "/" + intface + ".java");
			pw = new PrintWriter(new BufferedWriter(fw));
			// Pass in set of methods and get import classes
			DeclarationHandler decHandler = mapIntDeclHand.get(intface);
			InterfaceDecl intDecl = (InterfaceDecl) decHandler.getInterfaceDecl(intface);
			List<String> methods = intDecl.getMethods();
			Set<String> importClasses = getImportClasses(methods, intDecl);
			printImportStatements(importClasses);
			// Write interface header
			println("");
			println("public interface " + intface + " {");
			// Write methods
			for (String method : methods) {

				List<String> methParams = intDecl.getMethodParams(method);
				List<String> methPrmTypes = intDecl.getMethodParamTypes(method);
				print("public " + intDecl.getMethodType(method) + " " +
					intDecl.getMethodId(method) + "(");
				for (int i = 0; i < methParams.size(); i++) {
					// Check for params with driver class types and exchange it 
					// 		with its remote interface
					String paramType = checkAndGetParamClass(methPrmTypes.get(i));
					print(paramType + " " + methParams.get(i));
					// Check if this is the last element (don't print a comma)
					if (i != methParams.size() - 1) {
						print(", ");
					}
				}
				println(");");
			}
			println("}");
			pw.close();
			System.out.println("IoTCompiler: Generated local interface " + intface + ".java...");
		}
	}


	/**
	 * generateCplusLocalInterfaces() writes the local interfaces and provides type-checking.
	 * <p>
	 * It needs to rewrite and exchange USERDEFINED types in input parameters of stub
	 * and original interfaces, e.g. exchange Camera and CameraWithVideoAndRecording.
	 * The local interface has to be the input parameter for the stub and the stub 
	 * interface has to be the input parameter for the local class.
	 */
	public void generateCplusLocalInterfaces() throws IOException {

		// Create a new directory
		createDirectory(dir);
		for (String intface : mapIntfacePTH.keySet()) {
			// Open a new file to write into
			FileWriter fw = new FileWriter(dir + "/" + intface + ".hpp");
			pw = new PrintWriter(new BufferedWriter(fw));
			// Write file headers
			println("#include <iostream>");
			// Pass in set of methods and get include classes
			DeclarationHandler decHandler = mapIntDeclHand.get(intface);
			InterfaceDecl intDecl = (InterfaceDecl) decHandler.getInterfaceDecl(intface);
			List<String> methods = intDecl.getMethods();
			Set<String> includeClasses = getIncludeClasses(methods, intDecl);
			printIncludeStatements(includeClasses);
			println("");
			println("using namespace std;");
			println("");
			println("class " + intface);
			println("{");
			println("public:");
			// Write methods
			for (String method : methods) {

				List<String> methParams = intDecl.getMethodParams(method);
				List<String> methPrmTypes = intDecl.getMethodParamTypes(method);
				print("virtual " + convertType(intDecl.getMethodType(method)) + " " +
					intDecl.getMethodId(method) + "(");
				for (int i = 0; i < methParams.size(); i++) {
					// Check for params with driver class types and exchange it 
					// 		with its remote interface
					String paramType = checkAndGetParamClass(methPrmTypes.get(i));
					paramType = checkAndGetCplusType(paramType);
					print(paramType + " " + methParams.get(i));
					// Check if this is the last element (don't print a comma)
					if (i != methParams.size() - 1) {
						print(", ");
					}
				}
				println(") = 0;");
			}
			print("}");
			println(";");
			pw.close();
			System.out.println("IoTCompiler: Generated local interface " + intface + ".hpp...");
		}
	}


	/**
	 * generateJavaInterfaces() generate stub interfaces based on the methods list in Java
	 */
	public void generateJavaInterfaces() throws IOException {

		// Create a new directory
		String path = createDirectories(dir, subdir);
		for (String intface : mapIntfacePTH.keySet()) {

			Map<String,Set<String>> mapNewIntMethods = mapInt2NewInts.get(intface);
			for (Map.Entry<String,Set<String>> intMeth : mapNewIntMethods.entrySet()) {

				// Open a new file to write into
				String newIntface = intMeth.getKey();
				FileWriter fw = new FileWriter(path + "/" + newIntface + ".java");
				pw = new PrintWriter(new BufferedWriter(fw));
				DeclarationHandler decHandler = mapIntDeclHand.get(intface);
				InterfaceDecl intDecl = (InterfaceDecl) decHandler.getInterfaceDecl(intface);
				// Pass in set of methods and get import classes
				Set<String> importClasses = getImportClasses(intMeth.getValue(), intDecl);
				printImportStatements(importClasses);
				// Write interface header
				println("");
				println("public interface " + newIntface + " {");
				List<String> meths = intDecl.getMethods();
				// Write methods
				for (String method : intMeth.getValue()) {

					List<String> methParams = intDecl.getMethodParams(method);
					List<String> methPrmTypes = intDecl.getMethodParamTypes(method);
					print("public " + intDecl.getMethodType(method) + " " +
						intDecl.getMethodId(method) + "(");
					for (int i = 0; i < methParams.size(); i++) {
						print(methPrmTypes.get(i) + " " + methParams.get(i));
						// Check if this is the last element (don't print a comma)
						if (i != methParams.size() - 1) {
							print(", ");
						}
					}
					println(");");
				}
				println("}");
				pw.close();
				System.out.println("IoTCompiler: Generated interface " + newIntface + ".java...");
			}
		}
	}


	/**
	 * generateCPlusInterfaces() generate stub interfaces based on the methods list in C++
	 * <p>
	 * For C++ we use virtual classe as interface
	 */
	public void generateCPlusInterfaces() throws IOException {

		// Create a new directory
		String path = createDirectories(dir, subdir);
		for (String intface : mapIntfacePTH.keySet()) {

			Map<String,Set<String>> mapNewIntMethods = mapInt2NewInts.get(intface);
			for (Map.Entry<String,Set<String>> intMeth : mapNewIntMethods.entrySet()) {

				// Open a new file to write into
				String newIntface = intMeth.getKey();
				FileWriter fw = new FileWriter(path + "/" + newIntface + ".hpp");
				pw = new PrintWriter(new BufferedWriter(fw));
				DeclarationHandler decHandler = mapIntDeclHand.get(intface);
				InterfaceDecl intDecl = (InterfaceDecl) decHandler.getInterfaceDecl(intface);
				// Write file headers
				println("#include <iostream>");
				// Pass in set of methods and get import classes
				Set<String> includeClasses = getIncludeClasses(intMeth.getValue(), intDecl);
				printIncludeStatements(includeClasses);
				println("");
				println("using namespace std;");
				println("");
				println("class " + newIntface);
				println("{");
				println("public:");
				// Write methods
				for (String method : intMeth.getValue()) {

					List<String> methParams = intDecl.getMethodParams(method);
					List<String> methPrmTypes = intDecl.getMethodParamTypes(method);
					print("virtual " + convertType(intDecl.getMethodType(method)) + " " +
						intDecl.getMethodId(method) + "(");
					for (int i = 0; i < methParams.size(); i++) {

						String methPrmType = checkAndGetCplusType(methPrmTypes.get(i));
						print(methPrmType + " " + methParams.get(i));
						// Check if this is the last element (don't print a comma)
						if (i != methParams.size() - 1) {
							print(", ");
						}
					}
					println(") = 0;");
				}
				print("}");
				println(";");
				pw.close();
				System.out.println("IoTCompiler: Generated interface " + newIntface + ".hpp...");
			}
		}
	}


	/**
	 * generateJavaStubClasses() generate stubs based on the methods list in Java
	 */
	public void generateJavaStubClasses() throws IOException {

		// Create a new directory
		String path = createDirectories(dir, subdir);
		for (String intface : mapIntfacePTH.keySet()) {

			Map<String,Set<String>> mapNewIntMethods = mapInt2NewInts.get(intface);
			for (Map.Entry<String,Set<String>> intMeth : mapNewIntMethods.entrySet()) {

				// Open a new file to write into
				String newIntface = intMeth.getKey();
				String newStubClass = newIntface + "_Stub";
				FileWriter fw = new FileWriter(path + "/" + newStubClass + ".java");
				pw = new PrintWriter(new BufferedWriter(fw));
				DeclarationHandler decHandler = mapIntDeclHand.get(intface);
				InterfaceDecl intDecl = (InterfaceDecl) decHandler.getInterfaceDecl(intface);
				// Pass in set of methods and get import classes
				Set<String> importClasses = getImportClasses(intMeth.getValue(), intDecl);
				printImportStatements(importClasses);
				// Write interface header
				println("");
				println("public class " + newStubClass + " implements " + newIntface + " {");
				println("");
				// Write methods
				for (String method : intMeth.getValue()) {

					List<String> methParams = intDecl.getMethodParams(method);
					List<String> methPrmTypes = intDecl.getMethodParamTypes(method);
					print("public " + intDecl.getMethodType(method) + " " +
						intDecl.getMethodId(method) + "(");
					for (int i = 0; i < methParams.size(); i++) {

						print(methPrmTypes.get(i) + " " + methParams.get(i));
						// Check if this is the last element (don't print a comma)
						if (i != methParams.size() - 1) {
							print(", ");
						}
					}
					println(") {");
					// Check if this is not "void"
					if (!intDecl.getMethodType(method).equals("void")) {
						String retStmt = generateReturnStmt(intDecl.getMethodType(method));
						println("return " + retStmt + ";");
					}
					println("}"); println("");
				}
				println("}");
				pw.close();
				System.out.println("IoTCompiler: Generated stub class " + newStubClass + ".java...");
			}
		}
	}


	/**
	 * generateCPlusStubClasses() generate stubs based on the methods list in C++
	 */
	public void generateCPlusStubClasses() throws IOException {

		// Create a new directory
		String path = createDirectories(dir, subdir);
		for (String intface : mapIntfacePTH.keySet()) {

			Map<String,Set<String>> mapNewIntMethods = mapInt2NewInts.get(intface);
			for (Map.Entry<String,Set<String>> intMeth : mapNewIntMethods.entrySet()) {
				// Open a new file to write into
				String newIntface = intMeth.getKey();
				String newStubClass = newIntface + "_Stub";
				FileWriter fw = new FileWriter(path + "/" + newStubClass + ".hpp");
				pw = new PrintWriter(new BufferedWriter(fw));
				// Write file headers
				println("#include <iostream>");
				println("#include \"" + newIntface + ".hpp\""); println("");		
				println("using namespace std;"); println("");
				println("class " + newStubClass + " : public " + newIntface); println("{");
				println("public:"); println("");
				// Add default constructor and destructor
				println(newStubClass + "() { }"); println("");
				println("~" + newStubClass + "() { }"); println("");
				DeclarationHandler decHandler = mapIntDeclHand.get(intface);
				InterfaceDecl intDecl = (InterfaceDecl) decHandler.getInterfaceDecl(intface);
				// Write methods
				for (String method : intMeth.getValue()) {

					List<String> methParams = intDecl.getMethodParams(method);
					List<String> methPrmTypes = intDecl.getMethodParamTypes(method);
					print(convertType(intDecl.getMethodType(method)) + " " +
						intDecl.getMethodId(method) + "(");
					for (int i = 0; i < methParams.size(); i++) {

						String methPrmType = checkAndGetCplusType(methPrmTypes.get(i));
						print(methPrmType + " " + methParams.get(i));
						// Check if this is the last element (don't print a comma)
						if (i != methParams.size() - 1) {
							print(", ");
						}
					}
					println(") { ");
					// Check if this is not "void"
					if (!intDecl.getMethodType(method).equals("void")) {
						String retStmt = generateReturnStmt(intDecl.getMethodType(method));
						if (retStmt.equals("null")) { // null = NULL in C++
							retStmt = "NULL";
						}
						println("return " + retStmt + ";");
					}
					println("}"); println("");
				}
				print("}"); println(";");
				pw.close();
				System.out.println("IoTCompiler: Generated stub class " + newIntface + ".hpp...");
			}
		}
	}


	/**
	 * generateReturnStmt() generate return statement based on methType
	 */
	public String generateReturnStmt(String methType) {

		// Generate dummy returns for now
		if (methType.equals("short")||
			methType.equals("int") 	||
			methType.equals("long") ||
			methType.equals("float")||
			methType.equals("double")) {

			return "1";
		} else if ( methType.equals("String") ||
					methType.equals("byte")) {
  
			return "\"a\"";
		} else if ( methType.equals("char")) {

			return "\'a\'";
		} else if ( methType.equals("boolean")) {

			return "true";
		} else {
			return "null";
		}
	}


	/**
	 * setDirectory() sets a new directory for stub files
	 */
	public void setDirectory(String _subdir) {

		subdir = _subdir;
	}


	/**
	 * printUsage() prints the usage of this compiler
	 */
	public static void printUsage() {

		System.out.println();
		System.out.println("Sentinel interface and stub compiler version 1.0");
		System.out.println("Copyright (c) 2015-2016 University of California, Irvine - Programming Language Group.");
		System.out.println("All rights reserved.");
		System.out.println("Usage:");
		System.out.println("\tjava IoTCompiler -help / --help / -h\n");
		System.out.println("\t\tDisplay this help texts\n\n");
		System.out.println("\tjava IoTCompiler [<main-policy-file> <req-policy-file>]");
		System.out.println("\tjava IoTCompiler [<main-policy-file> <req-policy-file>] [options]\n");
		System.out.println("\t\tTake one or more pairs of main-req policy files, and generate Java and/or C++ files\n");
		System.out.println("Options:");
		System.out.println("\t-java\t<directory>\tGenerate Java stub files");
		System.out.println("\t-cplus\t<directory>\tGenerate C++ stub files");
		System.out.println();
	}


	/**
	 * parseFile() prepares Lexer and Parser objects, then parses the file
	 */
	public static ParseNode parseFile(String file) {

		ParseNode pn = null;
		try {
			ComplexSymbolFactory csf = new ComplexSymbolFactory();
			ScannerBuffer lexer = 
				new ScannerBuffer(new Lexer(new BufferedReader(new FileReader(file)),csf));
			Parser parse = new Parser(lexer,csf);
			pn = (ParseNode) parse.parse().value;
		} catch (Exception e) {
			e.printStackTrace();
			throw new Error("IoTCompiler: ERROR parsing policy file or wrong command line option: " + file);
		}

		return pn;
	}


	/**================
	 * Helper functions
	 **================
	 */
	boolean newline=true;
	int tablevel=0;

	private void print(String str) {
		if (newline) {
			int tab=tablevel;
			if (str.equals("}"))
				tab--;
			for(int i=0; i<tab; i++)
				pw.print("\t");
		}
		pw.print(str);
		updatetabbing(str);
		newline=false;
	}

	/**
	 * This function converts Java to C++ type for compilation
	 */
	private String convertType(String jType) {

		return mapPrimitives.get(jType);
	}


	private void println(String str) {
		if (newline) {
			int tab = tablevel;
			if (str.equals("}"))
				tab--;
			for(int i=0; i<tab; i++)
				pw.print("\t");
		}
		pw.println(str);
		updatetabbing(str);
		newline = true;
	}


	private void updatetabbing(String str) {
		tablevel+=count(str,'{')-count(str,'}');
	}


	private int count(String str, char key) {
		char[] array = str.toCharArray();
		int count = 0;
		for(int i=0; i<array.length; i++) {
			if (array[i] == key)
				count++;
		}
		return count;
	}


	private void createDirectory(String dirName) {

		File file = new File(dirName);
		if (!file.exists()) {
			if (file.mkdir()) {
				System.out.println("IoTCompiler: Directory " + dirName + " has been created!");
			} else {
				System.out.println("IoTCompiler: Failed to create directory " + dirName + "!");
			}
		} else {
			System.out.println("IoTCompiler: Directory " + dirName + " exists...");
		}
	}


	// Create a directory and possibly a sub directory
	private String createDirectories(String dir, String subdir) {

		String path = dir;
		createDirectory(path);
		if (subdir != null) {
			path = path + "/" + subdir;
			createDirectory(path);
		}
		return path;
	}


	// Inserting array members into a Map object
	// that maps arrKey to arrVal objects
	private void arraysToMap(Map map, Object[] arrKey, Object[] arrVal) {

		for(int i = 0; i < arrKey.length; i++) {

			map.put(arrKey[i], arrVal[i]);
		}
	}


	// Return parameter category, i.e. PRIMITIVES, NONPRIMITIVES, or USERDEFINED
	private ParamCategory getParamCategory(String paramType) {

		if (mapPrimitives.containsKey(paramType)) {
			return ParamCategory.PRIMITIVES;
		// We can either use mapNonPrimitivesJava or mapNonPrimitivesCplus here
		} else if (mapNonPrimitivesJava.containsKey(getSimpleType(paramType))) {
			return ParamCategory.NONPRIMITIVES;
		} else
			return ParamCategory.USERDEFINED;
	}


	// Return full class name for non-primitives to generate Java import statements
	// e.g. java.util.Set for Set, java.util.Map for Map
	private String getNonPrimitiveJavaClass(String paramNonPrimitives) {

		return mapNonPrimitivesJava.get(paramNonPrimitives);
	}


	// Return full class name for non-primitives to generate Cplus include statements
	// e.g. #include <set> for Set, #include <map> for Map
	private String getNonPrimitiveCplusClass(String paramNonPrimitives) {

		return mapNonPrimitivesCplus.get(paramNonPrimitives);
	}


	// Get simple types, e.g. HashSet for HashSet<...>
	// Basically strip off the "<...>"
	private String getSimpleType(String paramType) {

		// Check if this is generics
		if(paramType.contains("<")) {
			String[] type = paramType.split("<");
			return type[0];
		} else
			return paramType;
	}


	// Generate a set of classes for import statements
	private Set<String> getImportClasses(Collection<String> methods, InterfaceDecl intDecl) {

		Set<String> importClasses = new HashSet<String>();
		for (String method : methods) {
			List<String> methPrmTypes = intDecl.getMethodParamTypes(method);
			for (String paramType : methPrmTypes) {

				String simpleType = getSimpleType(paramType);
				if (getParamCategory(simpleType) == ParamCategory.NONPRIMITIVES) {
					importClasses.add(getNonPrimitiveJavaClass(simpleType));
				}
			}
		}
		return importClasses;
	}


	// Generate a set of classes for include statements
	private Set<String> getIncludeClasses(Collection<String> methods, InterfaceDecl intDecl) {

		Set<String> includeClasses = new HashSet<String>();
		for (String method : methods) {

			List<String> methPrmTypes = intDecl.getMethodParamTypes(method);
			for (String paramType : methPrmTypes) {

				String simpleType = getSimpleType(paramType);
				if (getParamCategory(simpleType) == ParamCategory.NONPRIMITIVES) {
					includeClasses.add(getNonPrimitiveCplusClass(simpleType));
				}
			}
		}
		return includeClasses;
	}


	private void printImportStatements(Set<String> importClasses) {

		for(String cls : importClasses) {
			println("import " + cls + ";");
		}
	}


	private void printIncludeStatements(Set<String> includeClasses) {

		for(String cls : includeClasses) {
			println("#include <" + cls + ">");
		}
	}


	// Get the C++ version of a non-primitive type
	// e.g. set for Set and map for Map
	// Input nonPrimitiveType has to be generics in format
	private String[] getTypeOfGeneric(String nonPrimitiveType) {

		// Handle <, >, and , for 2-type generic/template
		String[] substr = nonPrimitiveType.split("<")[1].split(">")[0].split(",");
		return substr;
	}


	private String checkAndGetCplusType(String paramType) {

		if (getParamCategory(paramType) == ParamCategory.PRIMITIVES) {
			return convertType(paramType);
		} else if (getParamCategory(paramType) == ParamCategory.NONPRIMITIVES) {

			// Check for generic/template format
			if (paramType.contains("<") && paramType.contains(">")) {

				String genericClass = getSimpleType(paramType);
				String[] genericType = getTypeOfGeneric(paramType);
				String cplusTemplate = null;
				if (genericType.length == 1) // Generic/template with one type
					cplusTemplate = getNonPrimitiveCplusClass(genericClass) + 
						"<" + convertType(genericType[0]) + ">";
				else // Generic/template with two types
					cplusTemplate = getNonPrimitiveCplusClass(genericClass) + 
						"<" + convertType(genericType[0]) + "," + convertType(genericType[1]) + ">";
				return cplusTemplate;
			} else
				return getNonPrimitiveCplusClass(paramType);
		} else
			// Just return it as is if it's not non-primitives
			return paramType;
	}


	// Get simple types, e.g. HashSet for HashSet<...>
	// Basically strip off the "<...>"
	private String checkAndGetParamClass(String paramType) {

		// Check if this is generics
		if(getParamCategory(paramType) == ParamCategory.USERDEFINED) {
			return exchangeParamType(paramType);
		} else
			return paramType;
	}


	// Returns the other interface for type-checking purposes for USERDEFINED
	//		classes based on the information provided in multiple policy files
	// e.g. return CameraWithXXX instead of Camera
	private String exchangeParamType(String intface) {

		// Param type that's passed is the interface name we need to look for
		//		in the map of interfaces, based on available policy files.
		DeclarationHandler decHandler = mapIntDeclHand.get(intface);
		if (decHandler != null) {
		// We've found the required interface policy files
			RequiresDecl reqDecl = (RequiresDecl) decHandler.getRequiresDecl(intface);
			Set<String> setExchInt = reqDecl.getInterfaces();
			if (setExchInt.size() == 1) {
				Iterator iter = setExchInt.iterator();
				return (String) iter.next();
			} else {
				throw new Error("IoTCompiler: Ambiguous stub interfaces: " + setExchInt.toString() + 
					". Only one new interface can be declared if the object " + intface +
					" needs to be passed in as an input parameter!");
			}
		} else {
		// NULL value - this means policy files missing
			throw new Error("IoTCompiler: Parameter type lookup failed for " + intface +
				"... Please provide the necessary policy files for user-defined types.");
		}
	}


	public static void main(String[] args) throws Exception {

		// If there is no argument or just "--help" or "-h", then invoke printUsage()
		if ((args[0].equals("-help") ||
			 args[0].equals("--help")||
			 args[0].equals("-h"))   ||
			(args.length == 0)) {

			IoTCompiler.printUsage();

		} else if (args.length > 1) {

			IoTCompiler comp = new IoTCompiler();
			int i = 0;				
			do {
				// Parse main policy file
				ParseNode pnPol = IoTCompiler.parseFile(args[i]);
				// Parse "requires" policy file
				ParseNode pnReq = IoTCompiler.parseFile(args[i+1]);
				// Get interface name
				String intface = ParseTreeHandler.getOrigIntface(pnPol);
				comp.setParseTree(intface, pnPol, pnReq);
				comp.getMethodsForIntface(intface);
				i = i + 2;
			// 1) Check if this is the last option before "-java" or "-cplus"
			// 2) Check if this is really the last option (no "-java" or "-cplus")
			} while(!args[i].equals("-java") &&
					!args[i].equals("-cplus") &&
					(i < args.length));

			// Generate everything if we don't see "-java" or "-cplus"
			if (i == args.length) {
				comp.generateJavaLocalInterfaces();
				comp.generateJavaInterfaces();
				comp.generateJavaStubClasses();
				comp.generateCplusLocalInterfaces();
				comp.generateCPlusInterfaces();
				comp.generateCPlusStubClasses();
			} else {
			// Check other options
				while(i < args.length) {
					// Error checking
					if (!args[i].equals("-java") &&
						!args[i].equals("-cplus")) {
						throw new Error("IoTCompiler: ERROR - unrecognized command line option: " + args[i]);
					} else {
						if (i + 1 < args.length) {
							comp.setDirectory(args[i+1]);
						} else
							throw new Error("IoTCompiler: ERROR - please provide <directory> after option: " + args[i]);

						if (args[i].equals("-java")) {
							comp.generateJavaLocalInterfaces();
							comp.generateJavaInterfaces();
							comp.generateJavaStubClasses();
						} else {
							comp.generateCplusLocalInterfaces();
							comp.generateCPlusInterfaces();
							comp.generateCPlusStubClasses();
						}
					}
					i = i + 2;
				}
			}
		} else {
		// Need to at least have exactly 2 parameters, i.e. main policy file and requires file
			IoTCompiler.printUsage();
			throw new Error("IoTCompiler: At least two arguments (main and requires policy files) have to be provided!");
		}
	}
}




