package iotrmi.Java;

import java.io.IOException;
import java.nio.ByteBuffer;
import java.util.Arrays;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.lang.reflect.Method;

import java.util.HashSet;
import java.util.Set;


/** Class IoTRMICall is a class that serves method calls on stub.
 *  <p>
 *  A stub will use an object of this class to send the method
 *  information, e.g. object identifier, method identifier, and
 *  parameters.
 *
 * @author      Rahmadi Trimananda <rtrimana @ uci.edu>
 * @version     1.0
 * @since       2016-10-04
 */
public class IoTRMICall {


	/**
	 * Class Properties
	 */
	private IoTRMIUtil rmiUtil;
	private IoTSocketClient rmiClient;


	/**
	 * Constructors
	 */
	public IoTRMICall(int _port, String _address, int _rev) throws IOException {

		rmiUtil = new IoTRMIUtil();
		rmiClient = new IoTSocketClient(_port, _address, _rev);
	}


	/**
	 * remoteCall() calls a method remotely by passing in parameters and getting a return Object
	 */
	public Object remoteCall(String methodSign, Class<?> retType, Class<?> retGenTypeKey, 
			Class<?> retGenTypeVal, Class<?>[] paramCls, Object[] paramObj) {

		// Send method info
		byte[] methodBytes = methodToBytes(methodSign, paramCls, paramObj);
		try {
			rmiClient.sendBytes(methodBytes);
		} catch (IOException ex) {
			ex.printStackTrace();
			throw new Error("IoTRMICall: Error when sending bytes - rmiClient.sendBytes()");
		}
		// Receive return value and return it to caller
		Object retObj = null;
		if (retType != void.class) {
			byte[] retObjBytes = null;
			try {
				retObjBytes = rmiClient.receiveBytes(retObjBytes);
			} catch (IOException ex) {
				ex.printStackTrace();
				throw new Error("IoTRMICall: Error when receiving bytes - rmiClient.receiveBytes()");
			}
			retObj = IoTRMIUtil.getParamObject(retType, retGenTypeKey, retGenTypeVal, retObjBytes);
		}
		return retObj;
	}


	/**
	 * methodToBytes() returns byte representation of a method
	 */
	public byte[] methodToBytes(String methodSign, Class<?>[] paramCls, Object[] paramObj) {

		// Get method ID in bytes
		byte[] methodId = IoTRMIUtil.getHashCodeBytes(methodSign);

		// Get byte arrays and calculate method bytes length
		int numbParam = paramObj.length;
		int methodLen = IoTRMIUtil.METHOD_ID_LEN;	// Initialized to the length of method ID
		byte[][] objBytesArr = new byte[numbParam][];
		for (int i=0; i < numbParam; i++) {
			// Get byte arrays for the objects
			objBytesArr[i] = IoTRMIUtil.getObjectBytes(paramObj[i]);
			String clsName = paramCls[i].getSimpleName();
			int paramLen = rmiUtil.getTypeSize(clsName);
			if (paramLen == -1) { 		// indefinite length
				methodLen = methodLen + IoTRMIUtil.PARAM_LEN;
			}
			methodLen = methodLen + objBytesArr[i].length;
		}

		// Construct method in byte array
		byte[] method = new byte[methodLen];
		int pos = 0;
		System.arraycopy(methodId, 0, method, 0, methodId.length);
		pos = pos + IoTRMIUtil.METHOD_ID_LEN;
		// Second iteration for copying bytes
		for (int i=0; i < numbParam; i++) {

			String clsName = paramCls[i].getSimpleName();
			int paramLen = rmiUtil.getTypeSize(clsName);
			if (paramLen == -1) { 		// indefinite length
				paramLen = objBytesArr[i].length;
				byte[] paramLenBytes = IoTRMIUtil.intToByteArray(paramLen);
				System.arraycopy(paramLenBytes, 0, method, pos, IoTRMIUtil.PARAM_LEN);
				pos = pos + IoTRMIUtil.PARAM_LEN;
			}		
			System.arraycopy(objBytesArr[i], 0, method, pos, paramLen);
			pos = pos + paramLen;
		}

		return method;
	}

	
	public static void main(String[] args) throws Exception {

		String[] test = { "123", "456", "789" };
		byte[] b = IoTRMIUtil.getObjectBytes(test);

		Boolean[] test2 = new Boolean[] { true, false, false };
		byte[] b2 = IoTRMIUtil.getObjectBytes(test2);

		System.out.println(Arrays.toString(b));
		System.out.println(Arrays.toString(b2));

		String[] c = (String[]) IoTRMIUtil.getParamObjectArray(String[].class, b);
		System.out.println(Arrays.toString(c));

		Boolean[] c2 = (Boolean[]) IoTRMIUtil.getParamObjectArray(Boolean[].class, b2);
		System.out.println(Arrays.toString(c2));

		// Set
		/*Set<String> set = new HashSet<String>();
		set.add("1234");
		set.add("5678");

		byte[] objBytes = IoTRMIUtil.getObjectBytes(set);
		System.out.println(Arrays.toString(objBytes));
		Object obj = IoTRMIUtil.getParamObject(Set.class, null, String.class, objBytes);

		@SuppressWarnings("unchecked")
		Set<String> setStr = (Set<String>) obj;
		System.out.println("Set: " + setStr.toString());*/

		// List
		/*List<Long> list = new ArrayList<Long>();
		list.add(12345678l);
		list.add(23455432l);
		list.add(34566543l);

		byte[] objBytes = IoTRMIUtil.getObjectBytes(list);
		System.out.println(Arrays.toString(objBytes));
		Object obj = IoTRMIUtil.getParamObject(List.class, null, Long.class, objBytes);

		@SuppressWarnings("unchecked")
		List<Long> listStr = (List<Long>) obj;
		System.out.println("List: " + listStr.toString());*/

		// Map
		Map<Long,Integer> map = new HashMap<Long,Integer>();
		map.put(12345678l, 1234);
		map.put(23455432l, 5678);
		map.put(34566543l, 4321);

		byte[] objBytes = IoTRMIUtil.getObjectBytes(map);
		System.out.println(Arrays.toString(objBytes));
		Object obj = IoTRMIUtil.getParamObject(Map.class, Long.class, Integer.class, objBytes);

		map = (Map<Long,Integer>) obj;
		System.out.println("Received map: " + map.toString());

		//@SuppressWarnings("unchecked")
		//List<Long> listStr = (List<Long>) obj;
		//System.out.println("List: " + listStr.toString());

	}
}
