package iotrmi.Java;

import java.io.IOException;
import java.nio.ByteBuffer;
import java.util.Arrays;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.lang.reflect.*;


/** Class IoTRMIObject is a class that stores info of an object.
 *  <p>
 *  It stores object ID, methods, method ID, method's signature 
 *  and parameters.
 *  This class also receive calls from different objects as they
 *  ask to execute certain methods remotely. This will have the 
 *  execution result (return value) sent back to 
 *
 * @author      Rahmadi Trimananda <rtrimana @ uci.edu>
 * @version     1.0
 * @since       2016-10-03
 */
public class IoTRMIObject {

	/**
	 * Class Properties
	 */
	private Map<Integer,String> mapHash2Sign;	// Map from hashcode(method ID) to signature
	private IoTRMIUtil rmiUtil;
	private IoTSocketServer rmiServer;
	private byte[] methodBytes;


	/**
	 * Constructors
	 */
	public IoTRMIObject(int _port, String[] _methodSign) throws  
		ClassNotFoundException, InstantiationException, 
			IllegalAccessException, IOException {

		rmiUtil = new IoTRMIUtil();
		mapHash2Sign = new HashMap<Integer,String>();
		methodBytes = null;
		getMethodIds(_methodSign);	// Initialize the method ID map
		rmiServer = new IoTSocketServer(_port);
		rmiServer.connect();
	}


	/**
	 * sendReturnObj() sends back return Object to client
	 */
	public void sendReturnObj(Object retObj) throws IOException {

		// Send back return value
		byte[] retObjBytes = IoTRMIUtil.getObjectBytes(retObj);
		rmiServer.sendBytes(retObjBytes);
	}


	/**
	 * getMethodBytes() waits for method transmission in bytes
	 */
	public void getMethodBytes() throws IOException {

		// Receive method info
		methodBytes = rmiServer.receiveBytes(methodBytes);
	}


	/**
	 * getSignature() gets method signature from bytes
	 */
	public String getSignature() {

		// Get method ID
		byte[] methodIdBytes = new byte[IoTRMIUtil.METHOD_ID_LEN];
		System.arraycopy(methodBytes, 0, methodIdBytes, 0, IoTRMIUtil.METHOD_ID_LEN);
		// Get Method object to handle method
		int methodId = IoTRMIUtil.byteArrayToInt(methodIdBytes);
		// Get method signature from the Map
		return mapHash2Sign.get(methodId);
	}


	/**
	 * getMethodParams() gets method params based on byte array received
	 * <p>
	 * Basically this is the format of a method in bytes:
	 * 1) 32-bit value of method ID (hash code)
	 * 2) m parameters with n-bit value each (m x n-bit)
	 * For the parameters that don't have definite length,
	 * we need to extract the length from a preceding 32-bit
	 * field in front of it.
	 *
	 * For primitive objects:
	 * | 32-bit method ID | m-bit actual data (fixed length)  |
	 * 
	 * For string, arrays, and non-primitive objects:
	 * | 32-bit method ID | 32-bit length | n-bit actual data | ...
	 * 
	 */
	public Object[] getMethodParams(Class<?>[] arrCls, Class<?>[] arrGenKeyCls, Class<?>[] arrGenValCls) {

		// Byte scanning position
		int pos = IoTRMIUtil.METHOD_ID_LEN;
		Object[] paramObj = new Object[arrCls.length];
		for (int i=0; i < arrCls.length; i++) {

			String paramType = arrCls[i].getSimpleName();
			int paramSize = rmiUtil.getTypeSize(paramType);
			// Get the 32-bit field in the byte array to get the actual
			// 		length (this is a param with indefinite length)
			if (paramSize == -1) {
				byte[] bytPrmLen = new byte[IoTRMIUtil.PARAM_LEN];
				System.arraycopy(methodBytes, pos, bytPrmLen, 0, IoTRMIUtil.PARAM_LEN);
				pos = pos + IoTRMIUtil.PARAM_LEN;
				paramSize = IoTRMIUtil.byteArrayToInt(bytPrmLen);
			}
			byte[] paramBytes = new byte[paramSize];
			System.arraycopy(methodBytes, pos, paramBytes, 0, paramSize);
			pos = pos + paramSize;
			paramObj[i] = IoTRMIUtil.getParamObject(arrCls[i], arrGenKeyCls[i], 
				arrGenValCls[i], paramBytes);
		}

		return paramObj;
	}
	

	/**
	 * getMethodIds() gets methods identifiers (hash code) and store them in the Map
	 */
	private void getMethodIds(String[] methodSign) {

		for (String sign : methodSign) {
			byte[] hashCode = IoTRMIUtil.getHashCodeBytes(sign);
			int methodId = IoTRMIUtil.byteArrayToInt(hashCode);
			mapHash2Sign.put(methodId, sign);
		}
	}
}
